
#include "base_fct.h"
#include "list_add.h"
#include "errors.h"

#define BASE_FCT_NB 31


enum {ARCCOSINE, ARCSINE, ARCTANGENT, COSINE, PLUS, MINUS, MULT, DIVIDE, CAR,
      CDR, CONS, EQ, EQUAL, EQBIS, EVAL, EMPTY, ATOM, LOWER, GREATER, LEQ, 
      GEQ, MIN, MAX, NEQ, PRINT, READ, SINE, TANGENT, LENGTH, SQRT, END};

struct base_fct_st{
  char name[8];
  list (*fct)(list arg);
};

typedef struct base_fct_st base_fct;

base_fct base_fct_tab[BASE_FCT_NB];

int is_cr_form(char *name)
{
  int i,l;
  l = strlen(name);
  if (name[0] != 'c') return FALSE;
  if (name[l-1] != 'r') return FALSE;
  if (l < 3) return FALSE;
  for(i = 1; i < l-2; i++)
    if ((name[i] != 'a') && (name[i] != 'd')) return FALSE;
  return TRUE;
}

list lisp_cr_form(char *name,list arg)
{
  list r;
  int l;
  int i;
  l = strlen(name);
  if (length(arg) != 1) return argnberror(name,1);
  r = car(arg);
  for(i = l-2; i > 0; i--) {
    if (is_empty(r)) {
      fprintf(stderr,"** %s : cannot take the %s of ",name,name);
      print_list(car(arg));
      printf("\n");
      raise(MYSIGERROR);
      }
    if (name[i] == 'a') r = car(r);
    else r = cdr(r);
  }
  return r;
}

list lisp_plus(list arg)
{
  if (is_empty(arg)) return num_to_list(0);
  if (!is_a_num(car(arg))) argtyperror("** + : not a number : ", car(arg));
  return num_to_list(list_to_num(car(arg)) + list_to_num(lisp_plus(cdr(arg))));
}

list lisp_minus(list arg)
{
  if (is_empty(arg)) return num_to_list(0);
  if (!is_a_num(car(arg))) argtyperror("** - : not a number : ", car(arg));
  if (length(arg) == 1) 
    return num_to_list(-list_to_num(car(arg)));
  return num_to_list(list_to_num(car(arg)) - list_to_num(lisp_plus(cdr(arg))));
}

list lisp_mult(list arg)
{
  if (is_empty(arg)) return num_to_list(1);
  if (!is_a_num(car(arg))) argtyperror("** * : not a number : ", car(arg));
  return num_to_list(list_to_num(car(arg)) * list_to_num(lisp_mult(cdr(arg))));
}

list lisp_divide(list arg)
{
  if (is_empty(arg)) return num_to_list(1);
  if (!is_a_num(car(arg))) argtyperror("** / : not a number : ", car(arg));
  if (length(arg) == 1)
    return num_to_list(1/list_to_num(car(arg)));
  return num_to_list(list_to_num(car(arg)) / list_to_num(lisp_mult(cdr(arg))));
}

list lisp_sine(list arg)
{ 
  if (length(arg) != 1) return argnberror("sin",1);
  return num_to_list(sin(list_to_num(car(arg))));
}


list lisp_tangent(list arg)
{ 
  if (length(arg) != 1) return argnberror("tan",1);
  return num_to_list(tan(list_to_num(car(arg))));
}

list lisp_cosine(list arg)
{ 
  if (length(arg) != 1) return argnberror("cos",1);
  return num_to_list(cos(list_to_num(car(arg))));
}

list lisp_arcsine(list arg)
{ 
  if (length(arg) != 1) return argnberror("asin",1);
  return num_to_list(asin(list_to_num(car(arg))));
}


list lisp_arctangent(list arg)
{ 
  if (length(arg) != 1) return argnberror("atan",1);
  return num_to_list(atan(list_to_num(car(arg))));
}

list lisp_arccosine(list arg)
{ 
  if (length(arg) != 1) return argnberror("acos",1);
  return num_to_list(acos(list_to_num(car(arg))));
}


list lisp_car(list arg)
{
  if (length(arg) != 1) return argnberror("car",1);
  return car(car(arg));
}

list lisp_cdr(list arg)
{
  if (length(arg) != 1) return argnberror("cdr",1);
  return cdr(car(arg));
}

list lisp_cons(list arg)
{
  if (length(arg) != 2) return argnberror("cons",2);
  return cons(car(arg),cadr(arg));
}

list lisp_eq(list arg)
{
  list l1,l2;
  if (length(arg) != 2) return argnberror("eq",2);
  l1 = car(arg);
  l2 = cadr(arg);
  if (is_empty(l1)) {
    if (is_empty(l1)) return list_true(); 
    else return list_false();
  }
  if (is_a_num(l1)) {
    if (is_a_num(l2)) {
      if (list_to_num(l1) == list_to_num(l2))
	return list_true();
      else return list_false();
    } 
    else return list_false();
  }
  if (is_a_name(l1)) {
    if (is_a_name(l2)) {
      if (strcmp(list_to_string(l1),list_to_string(l2)) == 0)
	return list_true();
      else return list_false();
    } 
    else return list_false();
  }
  if (is_a_cell(l2)) {
    if (l1 == l2) return list_true();
    else return list_false();
  }
  return list_false();
}

int list_equal(list l1, list l2)
{
  if (is_empty(l1)) {
    if (is_empty(l1))return TRUE; 
    else return FALSE;
  }
  if (is_a_num(l1)) {
    if (is_a_num(l2)) {
      if (list_to_num(l1) == list_to_num(l2))
	return TRUE;
      else return FALSE;
    } 
    else return FALSE;
  }
  if (is_a_name(l1)) {
    if (is_a_name(l2)) {
      if (strcmp(list_to_string(l1),list_to_string(l2)) == 0)
	return TRUE;
      else return FALSE;
    }
    else return FALSE;
  }
  if (is_a_cell(l2)) {
    if (list_equal(car(l1),car(l2)) == TRUE) {
      if (list_equal(cdr(l1),cdr(l2)) == TRUE)
	return TRUE;
      else return FALSE;
    }
    else return FALSE;
  }
  return FALSE;
}

list lisp_equal(list arg)
{
  if (length(arg) != 2) return argnberror("equal",2);
  if (list_equal(car(arg),cadr(arg)) == TRUE)
    return list_true();
  else return list_false();
}

list lisp_neq(list arg)
{
  if (length(arg) != 2) return argnberror("neq",2);
  if (list_equal(car(arg),cadr(arg)) == FALSE)
    return list_true();
  else return list_false();
}

list lisp_eval(list arg)
{
  if (length(arg) != 1)  { // && (length(arg) !=2)
    return argnberror("eval",1);
  }
  return eval_list(car(arg));
}

list lisp_null(list arg)
{
  if (length(arg) != 1) argnberror("null",1);
  if (is_empty(car(arg))) return list_true();
  return list_false();
}

list lisp_atom(list arg)
{
  if (length(arg) != 1) return argnberror("atom",1);
  if (is_empty(car(arg))) return list_true();
  if (is_a_num(car(arg))) return list_true();
  if (is_a_name(car(arg))) return list_true();
  return list_false();
}

list lisp_lower(list arg)
{
  if (length(arg) != 2) return argnberror("<",2);
  if (!is_a_num(car(arg))) argtyperror("** < : not a number : ", car(arg));
  if (!is_a_num(cadr(arg))) argtyperror("** < : not a number : ", cadr(arg));
  if (list_to_num(car(arg)) < list_to_num(cadr(arg))) return list_true();
  return list_false();
}

list lisp_greater(list arg)
{
  if (length(arg) != 2) return argnberror(">",2);
  if (!is_a_num(car(arg))) argtyperror("** > : not a number : ", car(arg));
  if (!is_a_num(cadr(arg))) argtyperror("** > : not a number : ", cadr(arg));
  if (list_to_num(car(arg)) > list_to_num(cadr(arg))) return list_true();
  return list_false();
}
list lisp_leq(list arg)
{
  if (length(arg) != 2) return argnberror("<=",2);
  if (!is_a_num(car(arg))) argtyperror("** <= : not a number : ", car(arg));
  if (!is_a_num(cadr(arg))) argtyperror("** <= : not a number : ", cadr(arg));
  if (list_to_num(car(arg)) <= list_to_num(cadr(arg))) return list_true();
  return list_false();
}

list lisp_geq(list arg)
{
  if (length(arg) != 2) return argnberror(">=",2);
  if (!is_a_num(car(arg))) argtyperror("** >= : not a number : ", car(arg));
  if (!is_a_num(cadr(arg))) argtyperror("** >= : not a number : ", cadr(arg));
  if (list_to_num(car(arg)) >= list_to_num(cadr(arg))) return list_true();
  return list_false();
}

list lisp_min(list arg)
{
  list t;
  if (length(arg) < 1) return argnberror("min",1);
  if (!is_a_num(car(arg))) argtyperror("** min : not a number : ", car(arg));
  if (length(arg) == 1) return car(arg);
  t = lisp_min(cdr(arg));
  if (list_to_num(car(arg)) < list_to_num(t)) return car(arg);
  return t;
}

list lisp_max(list arg)
{
  list t;
  if (length(arg) < 1) return argnberror("max",1);
  if (!is_a_num(car(arg))) argtyperror("** max : not a number : ", car(arg));
  if (length(arg) == 1) return car(arg);
  t = lisp_max(cdr(arg));
  if (list_to_num(car(arg)) > list_to_num(t)) return car(arg);
  return t;
}

list lisp_length(list arg)
{
  if (length(arg) != 1) return argnberror("length",1);
  return num_to_list(length(car(arg)));
}

list lisp_print(list arg)
{
  list last;
  list p = arg;
  last = empty_list();
  while (is_a_cell(p)) {
    last = car(p);
    print_list(last);
    p = cdr(p);
  }
  printf("\n");
  return last;
}

list lisp_read(list arg)   
{
  if (!is_empty(arg)) return argnberror("read",0);
  return read_list();
}

list lisp_end(list arg)
{
  raise(MYSIGEOF);
  return empty_list();
}

list lisp_sqrt(list arg)
{
  if (length(arg) != 1) return argnberror("srt",1);
  if (!is_a_num(car(arg)))
    return argtyperror("** sqrt : not a number : ",car(arg));
  return num_to_list(sqrt(list_to_num(car(arg))));
}

void init_base_fct()
{

  strcpy(base_fct_tab[PLUS].name,"+");
  base_fct_tab[PLUS].fct = &lisp_plus;

  strcpy(base_fct_tab[MINUS].name,"-");
  base_fct_tab[MINUS].fct = &lisp_minus;

  strcpy(base_fct_tab[MULT].name,"*");
  base_fct_tab[MULT].fct = &lisp_mult;

  strcpy(base_fct_tab[DIVIDE].name,"/");
  base_fct_tab[DIVIDE].fct = &lisp_divide;

  strcpy(base_fct_tab[CAR].name,"car");
  base_fct_tab[CAR].fct = &lisp_car;

  strcpy(base_fct_tab[CDR].name,"cdr");
  base_fct_tab[CDR].fct = &lisp_cdr;

  strcpy(base_fct_tab[CONS].name,"cons");
  base_fct_tab[CONS].fct = &lisp_cons;

  strcpy(base_fct_tab[EQ].name,"eq");
  base_fct_tab[EQ].fct = &lisp_eq;

  strcpy(base_fct_tab[EQUAL].name,"equal");
  base_fct_tab[EQUAL].fct = &lisp_equal;
  
  strcpy(base_fct_tab[EQBIS].name,"=");
  base_fct_tab[EQBIS].fct = &lisp_eq;

  strcpy(base_fct_tab[EVAL].name,"eval");
  base_fct_tab[EVAL].fct = &lisp_eval;

  strcpy(base_fct_tab[EMPTY].name,"null");
  base_fct_tab[EMPTY].fct = &lisp_null;

  strcpy(base_fct_tab[ATOM].name,"atom");
  base_fct_tab[ATOM].fct = &lisp_atom;

  // fonctions de comparaison  

  strcpy(base_fct_tab[LOWER].name,"<");
  base_fct_tab[LOWER].fct = &lisp_lower;

  strcpy(base_fct_tab[GREATER].name,">");
  base_fct_tab[GREATER].fct = &lisp_greater;

  strcpy(base_fct_tab[LEQ].name,"<=");
  base_fct_tab[LEQ].fct = &lisp_leq;

  strcpy(base_fct_tab[GEQ].name,">=");
  base_fct_tab[GEQ].fct = &lisp_geq;

  strcpy(base_fct_tab[NEQ].name,"neq");
  base_fct_tab[NEQ].fct = &lisp_neq;  

  strcpy(base_fct_tab[MIN].name,"min");
  base_fct_tab[MIN].fct = &lisp_min;

  strcpy(base_fct_tab[MAX].name,"max");
  base_fct_tab[MAX].fct = &lisp_max;
  
  strcpy(base_fct_tab[PRINT].name,"print");
  base_fct_tab[PRINT].fct = &lisp_print;
  
  strcpy(base_fct_tab[READ].name,"read");
  base_fct_tab[READ].fct = &lisp_read; 

  strcpy(base_fct_tab[END].name,"end");
  base_fct_tab[END].fct = &lisp_end;

    // fonctions trigo

  strcpy(base_fct_tab[SINE].name,"sin");
  base_fct_tab[SINE].fct = &lisp_sine; 

  strcpy(base_fct_tab[COSINE].name,"cos");
  base_fct_tab[COSINE].fct = &lisp_cosine;

  strcpy(base_fct_tab[TANGENT].name,"tan");
  base_fct_tab[TANGENT].fct = &lisp_tangent;

  strcpy(base_fct_tab[ARCSINE].name,"asin");
  base_fct_tab[ARCSINE].fct = &lisp_arcsine;

  strcpy(base_fct_tab[ARCCOSINE].name,"acos");
  base_fct_tab[ARCCOSINE].fct = &lisp_arccosine;

  strcpy(base_fct_tab[ARCTANGENT].name,"atan");
  base_fct_tab[ARCTANGENT].fct = &lisp_arctangent;

  strcpy(base_fct_tab[LENGTH].name,"length");
  base_fct_tab[LENGTH].fct = &lisp_length; 
  
  strcpy(base_fct_tab[SQRT].name,"sqrt");
  base_fct_tab[SQRT].fct = &lisp_sqrt;
}

list apply_base_function(char *name, list arg, environment env, int traced)
{
  int i;
  for(i = 0; i < BASE_FCT_NB; i++) {
    if (strcmp(base_fct_tab[i].name, name) == 0)
      return base_fct_tab[i].fct(arg);
  }
  if (is_cr_form(name)) return lisp_cr_form(name,arg);
  fprintf(stderr,"** eval : not a function : %s\n", name);
  raise(MYSIGERROR);
  return empty_list();
}
