#include "environment.h"
#include "list_add.h"
#include "errors.h"

environment create_env(int l)
{
  env_item *p;
  p = (env_item *) malloc(sizeof(env_item));
  assert(p != NULL);
  p->name = (char *) malloc(l*sizeof(char));
  assert(p->name != NULL);
  return p;
}

int is_bounded(char *name, environment local_env)
{
  environment p;
  p = local_env;   // On regarde d'abord dans les variables locales
  while (p != EMPTYENV) {
    if (strcmp(name,p->name) == 0) 
      return (p->traced)? 2 : 1; // pour viter le reparcours de tout l'environnement
    p = p->next;
  }
  p = it;   // On regarde ensuite dans les variables globales
  while (p != EMPTYENV) {
    if (strcmp(name,p->name) == 0) 
      return (p->traced)? 2 : 1; // pour viter le reparcours de tout l'environnement
    p = p->next;
  }
  return FALSE;
}

list get_value(char *name, environment local_env)
{
  environment p;
  p = local_env;
  if (strcmp(name,"t") == 0) return list_true();
  while (p != EMPTYENV) {
    if (strcmp(name,p->name) == 0) return p->value;
    p = p->next;
  }
  p = it;
  while (p != EMPTYENV) {
    if (strcmp(name,p->name) == 0) return p->value;
    p = p->next;
  }
  fprintf(stderr,"** eval : undefined variable : %s\n",name);
  raise(MYSIGERROR);
  return NULL;
}

void trace(char *name, environment local_env)
{
  environment p;
  p = local_env;
  if (strcmp(name,"t") == 0) return;
  while (p != EMPTYENV) {
    if (strcmp(name,p->name) == 0) { p->traced = TRUE; return;}
    p = p->next;
  }
  p = it;
  while (p != EMPTYENV) {
    if (strcmp(name,p->name) == 0) { p->traced = TRUE; return;}
    p = p->next;
  }
  argerror("** trace : Undefined variable : report bug to cmenier@ens-lyon.fr\n");
}

void untrace(char *name, environment local_env)
{
  environment p;
  p = local_env;
  if (strcmp(name,"t") == 0) return;
  while (p != EMPTYENV) {
    if (strcmp(name,p->name) == 0) { p->traced = FALSE; return; }
    p = p->next;
  }
  p = it;
  while (p != EMPTYENV) {
    if (strcmp(name,p->name) == 0) { p->traced = FALSE; return; }
    p = p->next;
  }
  argerror("** untrace : Undefined variable : report bug to cmenier@ens-lyon.fr\n");
}

environment set_value(char *name, list l, environment env)
{
  environment p;
  if (strcmp(name,"t") == 0) {
    (void) argerror("** set_value : not a variable : t\n");
    return EMPTYENV;
  }
  p = create_env(strlen(name));
  strcpy(p->name,name);
  p->value = l;   // Copie ncessaire ?
  p->next = env;
  p->traced = FALSE;
  return p;
}

environment set_list_value(list l, list v, environment env, int traced)
{
  environment p = env;
  list l_aux,v_aux;
  l_aux = l;
  v_aux = v;
  while ((is_a_cell(l_aux)) && (is_a_cell(v_aux))) {
    if (traced) {
      print_list(car(l_aux));
      printf(" = ");
      print_list(car(v_aux));
      printf(" , ");
    }
    p = set_value(list_to_string(car(l_aux)),car(v_aux),p);
    l_aux = cdr(l_aux);
    v_aux = cdr(v_aux);
  }
  if (traced) printf("\n");
  if (!is_empty(v_aux)) {
    (void) argerror("Too many arguments\n");
    return EMPTYENV;
  }
  if (!is_empty(l_aux)) {
    (void) argerror("Too few arguments\n");
    return EMPTYENV;
  }
  return p;
}

environment set_list_let_value(list l, environment env)
{
  environment p = env;
  list l_aux;
  l_aux = l;
  while (is_a_cell(l_aux)) {
    if (length(car(l_aux)) != 2) {
      (void) argtyperror("** let : wrong type of arguments : ",car(l_aux));
      return EMPTYENV;
    }
    if (!is_a_var(caar(l_aux))) {
      (void) argtyperror("** let : not a variable : ", caar(l_aux));
      return EMPTYENV;
    }
    p = set_value(list_to_string(car(car(l_aux))),eval(cadar(l_aux), env),p);
    l_aux = cdr(l_aux);
  }
  return p;
}

void free_env(int n, environment env)
{
  int i;
  environment p;
  p = env;
  for(i =0; i < n; i++)
    {
      environment old;
      if (p == EMPTYENV) break;
      old =p;
      p = p->next;
      free(old->name);
      free(old);
    }
}
