
#include "lisp.h"
#include "environment.h"
#include "errors.h"
#include "list_add.h"
#include "base_fct.h"
#include "keyword_fct.h"

environment it = EMPTYENV;

int traceall = FALSE;

int count;

void init_count()
{
  count = 0;
}

void incr()
{
  count++;
}

void decr()
{
  count--;
}

void print_string_level(char *s)
{
  int i;
  printf("\t");
  for(i = 0; i < count; i++)
    printf(" ");
  printf("%d",count);
  printf("%s",s);
}

list apply(list f, list arg, environment env, int traced);

list eval_wrapped(list l, environment env)
{
  if (is_empty(l) || is_a_num(l))
    return l;
  if (is_a_name(l)) 
    return get_value(list_to_string(l), env);
  if (is_a_cell(l)) {
    list p, args;
    p = car(l);
    if (is_a_name(p)) {
      char *s;
      s = list_to_string(p);
      if (strcmp(s,"apply") == 0) {
	if (length(l) != 3) return argnberror("apply",2);
     	return apply(cadr(l), eval(caddr(l), env), env, traceall);
      }
      if (is_keyword(s))
	return apply_keyword_fct(s, l, env, traceall);
    }
    args = map(cdr(l), eval, env);
    return apply(p, args, env, traceall);
  }
  return NULL;
}

list eval(list l, environment env)
{
  list result;
  if (traceall) {
    print_string_level(" --> ");
    print_list(l);
    printf("\n");
    incr();
  }
  result = eval_wrapped(l, env);
  if (traceall) {
    decr();
    print_string_level(" <-- ");
    print_list(result);
    printf("\n");
  }
  return result;
}

list apply(list f, list arg, environment env, int traced)
{
  int t;
  if (is_a_num(f)) return argerror("**eval : cannot apply a number\n");
  if (is_empty(f)) return argerror("**eval : cannot apply nil\n");
  if (is_a_name(f)) {
    if (t = is_bounded(list_to_string(f), env)) {
      list result;
      if ((t-1) || traced) { 
	print_list(f);
	printf("\t --> ");
      }
      result = apply(get_value(list_to_string(f), env), arg, env, (t-1) || traced);
      if ((t-1) || traced) {
	printf("\t<-- ");
	print_list(result);
	printf("\n");
      }
      return result;
    }
    return apply_base_function(list_to_string(f), arg, env, traced);
  }
  if (is_a_name(car(f))) {
    if (strcmp(list_to_string(car(f)), "lambda") == 0) {
      environment local_env;
      if (length(f) != 3) return argnberror("lambda",2);
      if (is_a_var(cadr(f))) {
	local_env = set_value(list_to_string(cadr(f)), arg, env);
	return eval(caddr(f), local_env);
      }
      else {
	local_env = set_list_value(cadr(f), arg, env, traced);
	return eval(caddr(f), local_env);
      }
    }
    return apply(eval(f, env), arg, env, traced);
  } 
  else 
    return apply(eval(f, env), arg, env, traced);
}

list eval_list(list l)
{
  init_count();
  return eval(l, EMPTYENV);
}
