#include "lisp.h"
#include "errors.h"
#include "list_add.h"
#include "environment.h"

#define KEYWORD_FCT_NB 14

enum {TRACEALL, TRACE, UNTRACE, QUOTE, IF, DEFINE, SETQ, LET, DE, LAMBDA,
      COND, WHILE, PROGN, FOR};

struct keyword_fct_st{
  char name[12];
  list (*fct)(list l, environment env);
};

typedef struct keyword_fct_st keyword_fct;

keyword_fct keyword_fct_tab[KEYWORD_FCT_NB];

char *lambda_str = "lambda";

list lisp_traceall(list l, environment env)
{
  traceall = TRUE;
  return empty_list();
}

list lisp_trace(list l, environment env)
{
  if (length(l) != 2) 
    return argnberror("trace",1);
  if (!is_a_var(cadr(l)))
    return argtyperror("** trace : not a variable : ", cadr(l));
  if (!is_bounded(list_to_string(cadr(l)), env))
    return argtyperror("** trace : undefined variable : ",cadr(l));
  trace(list_to_string(cadr(l)), env);
  return cadr(l);
}

list lisp_untrace(list l, environment env)
{
  if (length(l) != 2) 
    return argnberror("untrace",1);
  if (!is_a_var(cadr(l)))
    return argtyperror("** untrace : not a variable : ", cadr(l));
  if (!is_bounded(list_to_string(cadr(l)), env))
    return argtyperror("** untrace : undefined variable : ",cadr(l));
  untrace(list_to_string(cadr(l)), env);
  return cadr(l);
}

list lisp_quote(list l, environment env)
{
  if (length(cdr(l)) != 1) return argnberror("quote",1);
  return cadr(l);
}

list lisp_if(list l, environment env)
{
  if (length(cdr(l)) != 3) return argnberror("if",3);
  if (is_false(eval(cadr(l), env)))
    return eval(cadddr(l), env);
  return eval(caddr(l), env);
}

list lisp_define(list l, environment env)
{
  list p, args;
  if (length(l) != 3) return argnberror("define",2);
  p = cadr(l);
  if (!is_a_var(p)) argtyperror("** define : not a variable name : ", p);
  args = eval(caddr(l), env);
  it = set_value(list_to_string(p), args, it); 
  return args;
}

list lisp_let(list l, environment env)
{
  environment local_env;
  local_env = set_list_let_value(cadr(l), env);
  return eval(caddr(l), local_env);
}

list lisp_de(list l, environment env)
{
  list args;
  if (length(l) != 4) 
    return argnberror("de",3);
  if (!is_a_var(cadr(l))) 
    return argtyperror("** de : not a variable : ", cadr(l));
  args = caddr(l);
  while (is_a_cell(args)) {
    if (!is_a_var(car(args))) 
      return argtyperror("** de : bad parameter : ", car(args));
    args = cdr(args);
  }
  it = set_value(list_to_string(cadr(l)), cons(string_to_list(lambda_str),cddr(l)), it);
  return cadr(l);
}

list lisp_lambda(list l, environment env)
{
  return l;
}

list lisp_cond(list l, environment env)
{
  list args;
  args = cdr(l);
  while(is_a_cell(args)) {
    list c;
    c = car(args);
    if (length(c) != 2) return argnberror("cond : condition : ",2);
    if (!is_false(eval(car(c), env))) return eval(cadr(c), env);
    args = cdr(args);
  }
  return empty_list();
}

list lisp_while(list l, environment env)
{
  list cond;
  list body;
  list last = empty_list();
  if (length(cdr(l)) != 2) return argnberror(list_to_string(car(l)),2);
  cond = cadr(l);
  body = caddr(l);
  while (!is_false(eval(cond, env))) 
    last = eval(body, env);
  return last;
}

list lisp_progn(list l, environment env)
{
  list last;
  list args;
  last = empty_list();
  args = cdr(l);
  while(is_a_cell(args)) {
    last = eval(car(args), env);
    args = cdr(args);
  }
  return last;
}

list lisp_for(list l, environment env)
{
  list l1, l2;
  list last;
  list body;
  char *var;
  int i;
  environment local_env;
  if (length(l) != 5) return argnberror("for",4);
  if (!is_a_var(cadr(l))) 
    return argtyperror("** for : not a variable : ", cadr(l));
  var = list_to_string(cadr(l));
  l1 = eval(caddr(l), env);
  l2 = eval(cadddr(l), env);
  body = cadddr(cdr(l));
  if (!is_a_num(l1))
    return argtyperror("** for : not a number : ", l1);
  if (!is_a_num(l2))
    return argtyperror("** for : not a number : ", l2);
  last = empty_list();
  for(i = (int) list_to_num(l1); i <= (int) list_to_num(l2); i++)
    {
      local_env = set_value(var, num_to_list(i), env);
      last = eval(body, local_env);
    }
  return last;
}

int is_keyword(char *name)
{
  int i;
  for(i = 0; i < KEYWORD_FCT_NB; i++)
    if (strcmp(keyword_fct_tab[i].name, name) == 0) 
      return TRUE;
  return FALSE;
}

void init_keyword_fct()
{
  strcpy(keyword_fct_tab[TRACEALL].name,"traceall");
  keyword_fct_tab[TRACEALL].fct = &lisp_traceall;

  strcpy(keyword_fct_tab[TRACE].name,"trace");
  keyword_fct_tab[TRACE].fct = &lisp_trace;
 
  strcpy(keyword_fct_tab[UNTRACE].name,"untrace");
  keyword_fct_tab[UNTRACE].fct = &lisp_untrace;

  strcpy(keyword_fct_tab[QUOTE].name,"quote");
  keyword_fct_tab[QUOTE].fct = &lisp_quote; 

  strcpy(keyword_fct_tab[IF].name,"if");
  keyword_fct_tab[IF].fct = &lisp_if;

  strcpy(keyword_fct_tab[DEFINE].name,"define");
  keyword_fct_tab[DEFINE].fct = &lisp_define;

  strcpy(keyword_fct_tab[SETQ].name,"setq");
  keyword_fct_tab[SETQ].fct = &lisp_define;
 
  strcpy(keyword_fct_tab[LET].name,"let");
  keyword_fct_tab[LET].fct = &lisp_let;

  strcpy(keyword_fct_tab[DE].name,"de");
  keyword_fct_tab[DE].fct = &lisp_de; 

  strcpy(keyword_fct_tab[LAMBDA].name,"lambda");
  keyword_fct_tab[LAMBDA].fct = &lisp_lambda;

  strcpy(keyword_fct_tab[COND].name,"cond");
  keyword_fct_tab[COND].fct = &lisp_cond;
 
  strcpy(keyword_fct_tab[WHILE].name,"while");
  keyword_fct_tab[WHILE].fct = &lisp_while;

  strcpy(keyword_fct_tab[PROGN].name,"progn");
  keyword_fct_tab[PROGN].fct = &lisp_progn;

  strcpy(keyword_fct_tab[FOR].name,"for");
  keyword_fct_tab[FOR].fct = &lisp_for;

}

list apply_keyword_fct(char *name, list arg, environment env, int traced)
{
  int i;
  for(i = 0; i < KEYWORD_FCT_NB; i++)
    if (strcmp(keyword_fct_tab[i].name,name) == 0) 
      return keyword_fct_tab[i].fct(arg, env);
  return argerror("** eval : apply_keyword_fct : report bugs to cmenier@ens-lyon.fr\n");
}
