/* Copyright (c) 1998 by Lip Ens-Lyon.  Tous droits reverve's */

/***
   NOM
     read
   UTILISATION

   COMMENTAIRE

   HISTORIQUE
     bouge - Dec 7, 1998: Created.
***/

#ifndef __LISP
#define __LISP

#define MYSIGEOF SIGUSR1
#define MYSIGERROR SIGUSR2

#define TRUE 1
#define FALSE 0

#include <assert.h>

#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <ctype.h>
#include <strings.h>
#include <malloc.h>
#include <math.h>


enum type_token_enum {
    LPAR_TOKEN, RPAR_TOKEN, QUOTE_TOKEN, NUM_TOKEN, STRING_TOKEN
};

typedef enum type_token_enum type_token;

struct token_struct {
    type_token      the_type;
    union {
	double             as_num;
	char           *as_string;
    }               the_value;
};

typedef struct token_struct token;

token get_token();

token look_token();

enum type_object_enum {
    NIL_OBJECT, NUM_OBJECT, NAME_OBJECT, CELL_OBJECT
};

struct cell_struct;
struct object_struct;

typedef struct cell_struct cell;
typedef struct object_struct object;
typedef enum type_object_enum type_object;

typedef object *list;

/*
 * Warning: a "list" is a pointer to an object. A object may be either NIL,
 * or an INTeger, or a NAME (string), or a CELL made of two pointers to
 * objects (that is, lists, its car and its cdr). This is thus different from
 * the notion of a CAML list, where the car is an object and the cdr a list.
 */

struct cell_struct {
    object         *the_car;
    object         *the_cdr;
};

struct object_struct {
    type_object     the_type;
    union {
	double         as_num;
	char           *as_name;

	cell            as_cell;
    }               the_value;
};

struct env_item_struct;

typedef struct env_item_struct env_item;

typedef env_item * environment;

extern environment it;

extern int traceall;

extern list empty_list();

extern int  is_empty(list);

extern int  is_a_num(list);

extern int  is_a_name(list);

extern int  is_a_cell(list);

extern list num_to_list(double);

extern list string_to_list(char *);

extern list car(list);

extern list cdr(list);

extern list cons(list, list);

extern void print_newline();

extern void print_list(list);

extern list read_list();

extern list eval_list(list);

extern list eval(list, struct env_item_struct *);

double string_to_num(char *s);

#endif
