#include "lisp.h"


static object  *
alloc_new_object()
{
    object         *p;
    p = (object *) malloc(sizeof(object));
    assert(p != NULL);
    return p;
}

list empty_list()
{
    static object   obj;
    /* TRICK: To avoid replication! */
    obj.the_type = NIL_OBJECT;
    return &obj;
}

/*
 * Observe that the empty list is a pointer **real object**, which consumes
 * storage resources. An alternative implementation would be to make it a
 * NULL pointer, or maybe a ((void *)1) pointer to make it more explicit and
 * tractable. It is considered that having NIL as a real object makes things
 * more explicit... Well?
 */


int is_empty(list l)
{
    return (l->the_type == NIL_OBJECT);
}

int is_a_num(list l)
{
    return (l->the_type == NUM_OBJECT);
}

int is_a_name(list l)
{
    return (l->the_type == NAME_OBJECT);
}

int is_a_cell(list l)
{
    return (l->the_type == CELL_OBJECT);
}

list num_to_list(double n)
{
  object         *p;
    p = alloc_new_object();
    p->the_type = NUM_OBJECT;
    (p->the_value).as_num = n;
    return p;
}

list string_to_list(char *s)
{
    object         *p;
    p = alloc_new_object();
    p->the_type = NAME_OBJECT;
    (p->the_value).as_name = s;
    return p;
}

list car(list l)
{
    if (is_empty(l)) {
	fprintf(stderr, "Cannot take the car of nil!\n");
	raise(MYSIGERROR);
    }
    if ((l->the_type) != CELL_OBJECT) {
	fprintf(stderr, "Cannot take the car of a number or a name!\n");
	raise(MYSIGERROR);
    }
    return (l->the_value).as_cell.the_car;
}

list cdr(list l)
{
    if (is_empty(l))
	return (empty_list());
    if ((l->the_type) != CELL_OBJECT) {
	fprintf(stderr, "Cannot take the cdr of a number or a name!\n");
	raise(MYSIGERROR);
    }
    return (l->the_value).as_cell.the_cdr;
}


list cons(list x, list l)
{
    object         *p;
    p = alloc_new_object();

    p->the_type = CELL_OBJECT;
    (p->the_value).as_cell.the_car = x;
    (p->the_value).as_cell.the_cdr = l;

    return p;
}

double string_to_num(char *s)
{
  int i = 0;
  double r, a;
  int e;
  double sign = 1;
  double signexp = 1;
  r = 0;
  a = 1;
  e = 0;
  if (s[0] == '+') {i++; sign = 1;}
  else if (s[0] == '-') {i++; sign = -1;}
  while (s[i] != '\0') {
    if (s[i] == '.') {i++; break;}
    if ((s[i] == 'e') || (s[i] == 'E')) break;
    r = 10*r + s[i] - '0';
    i++;
  }
  while (s[i] != '\0') {
    a = 0.1 * a;
    if ((s[i] == 'e') || (s[i] == 'E')) {i++; break;}
    r = r + a * (s[i] - '0');
    i++;
  }
  if(s[i] != '0') {
    if (s[i] == '+') {i++; signexp = 1;}
    else if(s[i] == '-') {i++; signexp = -1;}
  }
  while (s[i] != '\0') {
    e = 10 * e + s[i] - '\0';
    i++;
  }
  return sign * r * pow(10, signexp * e);
}  
