/* Copyright (c) 1998 by Lip Ens-Lyon.  Tous droits reverve's */

/***
   NOM
     read
   UTILISATION

   COMMENTAIRE

   HISTORIQUE
     bouge - Dec 7, 1998: Created.
***/

#include "lisp.h"

list            read_list();

static          list
reverse_list(list l)
{
    list            acc;
    acc = empty_list();
    while (!is_empty(l)) {
	acc = cons(car(l), acc);
	l = cdr(l);
    }
    return acc;
}

/*
 * The grammar actually needs a lookahead of 1 token.  This is implemented by
 * a pair of look/get functions.  This could alternatively be simulated as
 * follows. The current token would be "pre-read" and passed to the
 * read_token_list(token) parser as a parameter. Then, the read_list topevel
 * call would merely be read_token_list(get_token()).
 */

list
read_list()
{
    token           t;
    t = get_token();

    switch (t.the_type) {
    case NUM_TOKEN:
	return num_to_list(t.the_value.as_num);
    case STRING_TOKEN:
	{
	    char           *s;
	    s = t.the_value.as_string;
	    if (strcasecmp(s, "nil") == 0)
		return empty_list();
	    else
		return string_to_list(t.the_value.as_string);
	}
    case RPAR_TOKEN:
	return read_list();
    case QUOTE_TOKEN:
	{
	    list            l;
	    l = read_list();
	    return cons(string_to_list("quote"),
			cons(l, empty_list()));
	}
    case LPAR_TOKEN:
	{
	    list            l;
	    l = empty_list();
	    for (;;) {

		if (look_token().the_type == RPAR_TOKEN) {
		    (void) get_token();
		    /* Just eat it! */
		    return reverse_list(l);
		    /*
		     * List l has actually been built from tail to head!
		     */
		}
		l = cons(read_list(), l);
	    }
	}
    default:
	assert(FALSE);

    }

}
